@@ -1,5 +1,5 @@
 #! /usr/bin/env python
-# -*- coding: UTF-8 -*-
+# -*- coding: utf-8 -*-
 import os
 import sys
 
@@ -15,15 +15,13 @@ def show_appended_python_code():
     return a
 
 if __name__ == '__main__':
-    out_dir = sys.argv[1] if len(sys.argv) > 1 else os.curdir
-    file_out = os.path.join(out_dir, 'append_solidpython_code.scad')
-
+    out_dir = sys.argv[1] if len(sys.argv) > 1 else None
     a = show_appended_python_code()
 
-    print("%(__file__)s: SCAD file written to: \n%(file_out)s" % vars())
     # ================================================================
     # = include_orig_code appends all python code as comments to the
     # = bottom of the generated OpenSCAD code, so the final document
     # = contains the easy-to-read python code as well as the SCAD.
     # = ------------------------------------------------------------ =
-    scad_render_to_file(a, file_out,  include_orig_code=True)
+    file_out = scad_render_to_file(a, out_dir=out_dir,  include_orig_code=True)
+    print(f"{__file__}: SCAD file written to: \n{file_out}")
